@@ -1,12 +1,16 @@
 
 #require(MASS)
 
-stmgplink = function(trainbed,testbed=NULL,gamma=1,taun=NULL,lambda=1,Z=NULL,Zte=NULL,plink="plink --noweb",maf=0.01,hwe=1e-4,geno=0.1,fout="stp",trainfout="train",testfout="test",ll=50,maxal=NULL,alc=NULL,tdir=NULL,Znames=NULL,trainphenofile=NULL,phenoname=NULL,pSum=NULL){
-	return( .STplink(trainbed=trainbed,testbed=testbed,gamma=gamma,lam=taun,ko=lambda,Z=Z,Zte=Zte,plink=plink,maf=maf,hwe=hwe,geno=geno,fout=fout,trainfout=trainfout,testfout=testfout,ll=ll,maxal=maxal,alc=alc,naive=F, omit.tail=F,Sgn=F,tdir=tdir,Znames=Znames,trainphenofile=trainphenofile,phenoname=phenoname,pSum=pSum) )
+
+
+#require(MASS)
+
+stmgplink = function(trainbed,testbed=NULL,gamma=1,taun=NULL,lambda=1,Z=NULL,Zte=NULL,plink="plink --noweb",maf=0.01,hwe=1e-4,geno=0.1,fout="stp",trainfout="train",testfout="test",ll=50,maxal=NULL,alc=NULL,tdir=NULL,Znames=NULL,trainphenofile=NULL,testphenofile=NULL,phenoname=NULL,pSum=NULL){
+	return( .STplink(trainbed=trainbed,testbed=testbed,gamma=gamma,lam=taun,ko=lambda,Z=Z,Zte=Zte,plink=plink,maf=maf,hwe=hwe,geno=geno,fout=fout,trainfout=trainfout,testfout=testfout,ll=ll,maxal=maxal,alc=alc,naive=F, omit.tail=F,Sgn=F,tdir=tdir,Znames=Znames,trainphenofile=trainphenofile,testphenofile=testphenofile,phenoname=phenoname,pSum=pSum) )
 }
 
 
-.STplink = function(trainbed,testbed=NULL,gamma=1,lam=NULL,ko=1,Z=NULL,Zte=NULL,plink="plink --noweb",maf=0.01,hwe=1e-4,geno=0.1,fout="stp",trainfout="train",testfout="test",ll=50,maxal=NULL,alc=NULL,naive=F, omit.tail=F, Sgn=F, Nrow=10000, tdir=NULL, Znames=NULL, trainphenofile=NULL, phenoname=NULL, pSum=NULL){
+.STplink = function(trainbed,testbed=NULL,gamma=1,lam=NULL,ko=1,Z=NULL,Zte=NULL,plink="plink --noweb",maf=0.01,hwe=1e-4,geno=0.1,fout="stp",trainfout="train",testfout="test",ll=50,maxal=NULL,alc=NULL,naive=F, omit.tail=F, Sgn=F, Nrow=10000, tdir=NULL, Znames=NULL, trainphenofile=NULL, testphenofile=NULL, phenoname=NULL, pSum=NULL){
   #trainbed="train"; gamma=1;lam=c(0.5);maxal=NULL;ll=50;Z=NULL;maf=0;r2=0;naive=T; ko=1; omit.tail=F; Sgn=F; tdir=NULL; pSum=NULL
 
 
@@ -34,6 +38,7 @@ stmgplink = function(trainbed,testbed=NULL,gamma=1,taun=NULL,lambda=1,Z=NULL,Zte
   #read.assoc.ch(fi="stp.qassoc",maxal=0.001,pSum=s)
 
 
+  #if(is.null(tdir)){ tdir = paste(c(tempdir(),"/",sample(c(0:9,letters,LETTERS),15)),collapse=""); dir.create(tdir)  }
   if(is.null(tdir)) tdir = tempdir()
   fout = paste(tdir,fout,sep="/")
   trainfout = paste(tdir,trainfout,sep="/")
@@ -47,8 +52,8 @@ stmgplink = function(trainbed,testbed=NULL,gamma=1,taun=NULL,lambda=1,Z=NULL,Zte
      Phen = read.table(trainphenofile,header=TRUE,na.strings="-9"); rownames(Phen) = as.character(paste(Phen[,1],Phen[,2]))
      cat(paste0("Reading phenotype ",phenoname," from ",trainphenofile,"\n"))
      FAM[,6] = NA; iii = intersect(rownames(FAM),rownames(Phen)); FAM[iii,6] = Phen[iii,phenoname]
-     trainbed[3] = paste0(trainfout,".tmp.fam")  #.
-     write.table(FAM,trainbed[3],row.names=FALSE,col.names=FALSE,quote=FALSE)  #.
+     trainbed[3] = paste0(trainfout,".tmp.fam")  #
+     write.table(FAM,trainbed[3],row.names=FALSE,col.names=FALSE,quote=FALSE)  #
   }
 
 
@@ -57,8 +62,7 @@ stmgplink = function(trainbed,testbed=NULL,gamma=1,taun=NULL,lambda=1,Z=NULL,Zte
       Z = read.table(Z,header=TRUE,na.strings="-9"); rownames(Z) = as.character(paste(Z[,1],Z[,2])); Z = as.matrix(Z[,-(1:2),drop=FALSE]);
       if(!is.null(Znames)){
          cat( paste0( "Used covariates: ",paste(intersect(colnames(Z),Znames),collapse=","),"\n" ) )
-         #Z = Z[,intersect(colnames(Z),Znames)]
-         Z = Z[,intersect(colnames(Z),Znames),drop=FALSE]  #.
+         Z = Z[,intersect(colnames(Z),Znames),drop=FALSE]
       }
     }
     stopifnot(nrow(Z)==nrow(FAM))
@@ -203,7 +207,7 @@ stmgplink = function(trainbed,testbed=NULL,gamma=1,taun=NULL,lambda=1,Z=NULL,Zte
   # Estimated predicted values at optimal al for training data
   if(!is.null(Z)){  # Covariates
     if(!allZero){
-      mustp = cbind(Z[rownames(FAM),],XG)%*%ST$BA[-1,lopt[1],lopt[2]] + outer(rep(1,nrow(XG)),ST$BA[1,lopt[1],lopt[2]])
+      mustp = cbind(Z[rownames(FAM),,drop=FALSE],XG)%*%ST$BA[-1,lopt[1],lopt[2]] + outer(rep(1,nrow(XG)),ST$BA[1,lopt[1],lopt[2]])
     }else{
       mustp = Z%*%ST$BA[1+(1:ncol(Z)),lopt[1],lopt[2]] + outer(rep(1,nrow(Z)),ST$BA[1,lopt[1],lopt[2]])
     }
@@ -218,6 +222,14 @@ stmgplink = function(trainbed,testbed=NULL,gamma=1,taun=NULL,lambda=1,Z=NULL,Zte
   if(!is.null(testbed)){
     FAMte = read.table(testbed[3]); rownames(FAMte) = paste(FAMte[,1],FAMte[,2])
 
+    if(!is.null(testphenofile)){
+       stopifnot(!is.null(phenoname))
+       Phen = read.table(testphenofile,header=TRUE,na.strings="-9"); rownames(Phen) = as.character(paste(Phen[,1],Phen[,2]))
+       cat(paste0("Reading phenotype ",phenoname," from ",testphenofile,"\n"))
+       FAMte[,6] = NA; iii = intersect(rownames(FAMte),rownames(Phen)); FAMte[iii,6] = Phen[iii,phenoname]
+    }
+
+
     if(!allZero){
       #XG0te = read.table(paste(testfout,".raw",sep=""),header=TRUE); XG0te[which(XG0te[,6]==-9),6] = NA; rownames(XG0te) = paste(XG0te[,1],XG0te[,2]); XG0te = XG0te[rownames(FAMte),]
       #XGte = XG0te[,-(1:6)]; XGte = (XGte==1)+(XGte==2)*2;
@@ -231,11 +243,10 @@ stmgplink = function(trainbed,testbed=NULL,gamma=1,taun=NULL,lambda=1,Z=NULL,Zte
     # Estimated predicted values at optimal al for test data
     if(!is.null(Zte)){  # Covariates
       if(is.character(Zte)){  # plink .cov format
-        Zte = read.table(Zte,header=TRUE,na.strings="-9"); rownames(Zte) = as.character(paste(Zte[,1],Zte[,2])); Zte = as.matrix(Zte[,-(1:2),drop=FALSE]); #
+        Zte = read.table(Zte,header=TRUE,na.strings="-9"); rownames(Zte) = as.character(paste(Zte[,1],Zte[,2])); Zte = as.matrix(Zte[,-(1:2),drop=FALSE]);  #
         if(!is.null(Znames)){
            cat( paste0( "Used covariates: ",paste(intersect(colnames(Zte),Znames),collapse=","),"\n" ) )
-           #Zte = Zte[,intersect(colnames(Zte),Znames)]
-           Zte = Zte[,intersect(colnames(Zte),Znames),drop=FALSE] #.
+           Zte = Zte[,intersect(colnames(Zte),Znames),drop=FALSE]  #
         }
       }
       stopifnot(nrow(Zte)==nrow(FAMte))
@@ -504,89 +515,383 @@ return(list(Loss=Loss[-1,,drop=FALSE],Muhat=Muhat[,-1,,drop=FALSE],gdf=gdf[-1,,d
 }
 
 
-## Test
-if(0){
 
-  D = read.table("stp.raw",header=T)
-  X = D[,-(1:6)]
-  X = (X==1) + 2*(X==2)
-  p = ncol(X)
-  n = nrow(X)
-  R = 100; ll = 30
-  p0 = 10; b0 = log(rep(1.2,p0))
-  Op = Oe = GDF = matrix(0,R,ll)
-  qua = "q"
-  sig = 1.2
-  pSum = 1
-  crr = numeric(R)
+##########################################################################################
 
-  for(r in 1:R){
-    iA0 = sample(1:p,p0)
-    if(!is.null(pSum)){
-      pSum = runif(p); pSum[iA0] = pchisq(rnorm(length(b0),mean=4*exp(b0),sd=1)^2,df=1,lower.tail=F); pSum = matrix(pSum)
-    }
-    Z = as.matrix(cbind(rnorm(n),runif(n)))
-    eta = X[,iA0]%*%b0 - 4 + Z%*%c(0.1,0.1)/20
-    if(probit){
-      eta = eta - mean(eta)
-      mu = pnorm(eta*3-2.5)
-      y = rbinom(length(mu),size=1,prob=mu)
-      z = Z
-      x = X
-      N = n
-      ica = which(y==1)
-      ico = which(y==0)[1:length(ica)]
-      y = y[c(ica,ico)]
-      mu = mu[c(ica,ico)]
-      z = Z[c(ica,ico),]
-      x = X[c(ica,ico),]
-      N = length(y)
-    }else{
-      if(qua==""){
-        mu = 1/(1+exp(-eta))
-        #			mu = exp(eta); mu[which(mu>0.9)] = 0.9
-        y = rbinom(n,size=1,prob=mu)
-        z = Z
-        x = X
-        N = n
-        #			ica = which(y==1)
-        #			ico = which(y==0)[1:length(ica)]
-        #			y = y[c(ica,ico)]
-        #			mu = mu[c(ica,ico)]
-        #			z = Z[c(ica,ico),]
-        #			x = X[c(ica,ico),]
-        #			N = length(y)
-      }else{
-        mu = eta
-        y = mu + rnorm(n)*sig
-        z = Z
-        x = X
-        N = n
-      }
-    }
-    ST = .stee(y,x,z,Tau=N*c(1),qua,maxal=0.1,gamma=1,ll=ll,naive=F,pSum=pSum)
-    MU = outer(as.vector(mu),rep(1,ll))
-    Y = outer(as.vector(y),rep(1,ll))
-    if(qua==""){
-      Op[r,] = -2*colSums( MU*ST$Muhat[,,1] - log(1 + exp(ST$Muhat[,,1])) )
-      Oe[r,] = -2*colSums( Y*ST$Muhat[,,1] - log(1 + exp(ST$Muhat[,,1])) )
-      GDF[r,] = ST$gdf
-    }else{
-      Op[r,] = colSums( (MU - ST$Muhat[,,1])^2 ) + N*(sig^2)
-      Oe[r,] = colSums( (Y - ST$Muhat[,,1])^2  )
-      GDF[r,] = (sig^2)*ST$gdf
-    }
+.tapproxq = function(Z,X,Y){  # quantitative
+	n = length(Y)
+	lll = lm(Y~Z-1)
+	diagQz = 1 - hatvalues(lll)
+	QzY = lll$resid
 
-    crr[r] = cor(as.vector(mu),ST$Muhat[,which.min(Oe[r,]+2*GDF[r,]),1])
+	D = t(X) %*% (X*as.vector(diagQz))
+	A_2 = t(X)%*%( X * ( QzY^2 ) )
+
+	A_3 = sum( QzY^2 )
+
+	AN = sum(diag(solve(D)%*%A_2))
+	t_approx = AN
+
+	t_approx = t_approx / ((1/n)*(A_3 - AN))
+
+	return(t_approx)
+}
+
+
+.tapproxb = function(Z,X,Y){  # binary
+	n = length(Y)
+	g0 = glm(Y~Z,family=binomial)
+	mu = fitted(g0)
+	om = sqrt(mu*(1-mu))
+	Z = om*Z
+	X = om*X
+	Y = (Y - mu)/om
+
+	lll = lm(Y~Z-1)
+	diagQz = 1 - hatvalues(lll)
+	QzY = lll$resid
+
+	D = t(X) %*% (X*diagQz)
+	A_2 = t(X)%*%( X * ( QzY^2 ) )
+
+	A_3 = sum( QzY^2 )
+
+	t_approx  = sum(diag(solve(D)%*%A_2))
+
+	return(t_approx)
+}
+
+
+.check.lapprox.ge = function(Z,X,y){
+	ii = complete.cases(cbind(Z,X,y));
+	Y = y[ii]; Z = cbind(1,Z[ii,,drop=FALSE]); X = X[ii,,drop=FALSE]  # include intercept
+	cat("Z must include E\n")
+	kkk = ifelse(length(table(Y))==2,.tapproxb(Z,X,Y),.tapproxq(Z,X,Y))/ncol(X)
+	cat(paste("lapprox is",kkk,", if it is far from 1 null model can be missepcified\n"))
+	return(kkk)
+}
+
+
+
+.divplink.train.test = function(itrain,itest,bed,fout,di=".",covfile=NULL,phenfile=NULL,plink="plink1.9"){
+	if(length(bed)==1) bed = paste0(bed,c(".bed",".bim",".fam"))
+	Fa = read.table(bed[3],header=FALSE); rownames(Fa) = paste(Fa[,1],Fa[,2])
+	write.table(Fa[itrain,],file=paste0(di,"/train",fout,".txt"),row.names=FALSE,col.names=FALSE,quote=FALSE)
+
+write.table(Fa[itest,],file=paste0(di,"/test",fout,".txt"),row.names=FALSE,col.names=FALSE,quote=FALSE)
+
+	system(paste0(plink," --bed ",bed[1]," --bim ",bed[2]," --fam ",bed[3]," --make-bed --out ",paste0(di,"/train",fout)," --keep ",paste0(di,"/train",fout,".txt")))
+	system(paste0(plink," --bed ",bed[1]," --bim ",bed[2]," --fam ",bed[3]," --make-bed --out ",paste0(di,"/test",fout)," --keep ",paste0(di,"/test",fout,".txt")))
+
+	if(!is.null(covfile)){
+		Co = read.table(covfile,header=TRUE); rownames(Co) = paste(Co[,1],Co[,2])
+		write.table(Co[intersect(rownames(Fa[itrain,]),rownames(Co)),],file=paste0(di,"/train",fout,".cov"),quote=FALSE,row.names=FALSE)
+		write.table(Co[intersect(rownames(Fa[itest,]),rownames(Co)),],file=paste0(di,"/test",fout,".cov"),quote=FALSE,row.names=FALSE)
+	}
+
+	if(!is.null(phenfile)){
+		Ph = read.table(phenfile,header=TRUE); rownames(Ph) = paste(Ph[,1],Ph[,2])
+		write.table(Ph[intersect(rownames(Fa[itrain,]),rownames(Ph)),],file=paste0(di,"/train",fout,".pheno"),quote=FALSE,row.names=FALSE)
+		write.table(Ph[intersect(rownames(Fa[itest,]),rownames(Ph)),],file=paste0(di,"/test",fout,".pheno"),quote=FALSE,row.names=FALSE)
+	}
+	
+}
+
+
+
+##########################################################################################
+
+
+stmgeplink = function(trainbed,Z,Enames,Zte=NULL,testbed=NULL,gamma=1,taun=NULL,lambda=1,plink="plink --noweb",maf=0.01,hwe=1e-4,geno=0.1,fout="stp",trainfout="train",testfout="test",ll=50,maxal=NULL,alc=NULL,tdir=NULL,Znames=NULL,trainphenofile=NULL,testphenofile=NULL,phenoname=NULL,Assc=FALSE,AsscGE=FALSE,centerE=TRUE){
+	return( .STgeplink(trainbed=trainbed,Z=Z,Enames=Enames,Zte=Zte,testbed=testbed,gamma=gamma,lam=taun,ko=lambda,plink=plink,maf=maf,hwe=hwe,geno=geno,fout=fout,trainfout=trainfout,testfout=testfout,ll=ll,maxal=maxal,alc=alc,naive=F, omit.tail=F,Sgn=F,tdir=tdir,Znames=Znames,trainphenofile=trainphenofile,testphenofile=testphenofile,phenoname=phenoname,Assc=Assc,AsscGE=AsscGE,centerE=centerE) )
+}
+
+
+.STgeplink = function(trainbed,Z,Enames,Zte=NULL,testbed=NULL,gamma=1,lam=NULL,ko=1,plink="plink --noweb",maf=0.01,hwe=1e-4,geno=0.1,fout="stp",trainfout="train",testfout="test",ll=50,maxal=NULL,alc=NULL,naive=F, omit.tail=F, Sgn=F, Nrow=10000, tdir=NULL, Znames=NULL, trainphenofile=NULL, testphenofile=NULL, phenoname=NULL,Assc=FALSE,AsscGE=FALSE,centerE=TRUE){
+  #trainbed="train"; gamma=1;lam=c(0.5);maxal=NULL;ll=50;Enames="COV1";Z="train.cov";maf=0;r2=0;naive=T; ko=1; omit.tail=F; Sgn=F; tdir=NULL; pSum=NULL; fout="stp";trainfout="train";testfout="test"; Znames=c("COV1","COV2"); trainphenofile=NULL; phenoname=NULL; plink="plink1.9"; maf=0.01;hwe=1e-4;geno=0.1; alc=NULL; testbed=NULL; library(MASS);
+
+
+  #if(is.null(tdir)){ tdir = paste(c(tempdir(),"/",sample(c(0:9,letters,LETTERS),15)),collapse=""); dir.create(tdir)  }
+  if(is.null(tdir)){ tdir = tempdir()  }
+  fout = paste(tdir,fout,sep="/")
+  trainfout = paste(tdir,trainfout,sep="/")
+  testfout = paste(tdir,testfout,sep="/")
+  if(length(trainbed)==1) trainbed = paste0(trainbed,c(".bed",".bim",".fam"))
+  nsnpE = length(readLines(trainbed[2]))*(1+length(Enames))
+  FAM = read.table(trainbed[3],na.strings="-9"); rownames(FAM) = as.character(paste(FAM[,1],FAM[,2]))
+
+  if(!is.null(trainphenofile)){
+    stopifnot(!is.null(phenoname))
+    Phen = read.table(trainphenofile,header=TRUE,na.strings="-9"); rownames(Phen) = as.character(paste(Phen[,1],Phen[,2]))
+    cat(paste0("Reading phenotype ",phenoname," from ",trainphenofile,"\n"))
+    FAM[,6] = NA; iii = intersect(rownames(FAM),rownames(Phen)); FAM[iii,6] = Phen[iii,phenoname]
+    trainbed[3] = paste0(trainfout,".tmp.fam")  #
+    write.table(FAM,trainbed[3],row.names=FALSE,col.names=FALSE,quote=FALSE)  #
+  }
+
+  stopifnot(!is.null(Enames))
+  stopifnot(!is.null(Z))
+
+
+  if(is.character(Z)){  # plink .cov format
+    Z = read.table(Z,header=TRUE,na.strings="-9"); rownames(Z) = as.character(paste(Z[,1],Z[,2])); Z = as.matrix(Z[,-(1:2),drop=FALSE]);
+    if(!is.null(Znames)){
+      cat( paste0( "Used covariates: ",paste(intersect(colnames(Z),Znames),collapse=","),"\n" ) )
+      Z = Z[,intersect(colnames(Z),Znames),drop=FALSE]
+    }
+  }
+  stopifnot(nrow(Z)==nrow(FAM))
+  Z = as.matrix(Z)
+  if(is.null(rownames(Z))){
+    cat("Assuming rownames of covariate Z are identical to those of .fam file\n")
+    rownames(Z) = as.character(paste(FAM[,1],FAM[,2]));
+  }
+  iZFAM = as.character( intersect(rownames(Z),rownames(FAM)) )  #
+  if(length(iZFAM)==0) cat("rownames of covariates (Z) need to be 'FID IID' in .fam file (FID[space]IID)")
+  cat(paste("number of individuals in covariate Z overlapping in .fam file is",length(iZFAM),"\n"))
+  FAM = FAM[iZFAM,]; Z = Z[iZFAM,,drop=FALSE]
+  ia = which(complete.cases(cbind(FAM,Z)))
+  FAM = FAM[ia,]; Z = Z[rownames(FAM),,drop=FALSE]
+
+  write.table(FAM,file=paste0(fout,".keep.fam"),row.names=F,col.names=F,quote=F)
+
+  y = as.numeric(FAM[,6]); names(y) = paste(FAM[,1],FAM[,2])
+  n = sum(!is.na(y))
+  nco = sum(y==1,na.rm=TRUE); nca = sum(y==2,na.rm=TRUE)
+  qua = ifelse(nco+nca==n,"","q")
+  if(qua=="") y = y-1
+  if(is.null(lam)){
+    Tau = n/sqrt(log(n))
+  }else{
+    Tau = n*lam
+  }
+  kk = ko
+
+  alo = ifelse(qua=="",3,3)*(n/log(n))/nsnpE  # used for sig2hato
+  if(is.null(maxal)) maxal = ifelse(qua=="",1,3)*alo
+
+
+  ZZZ = Z[as.character(paste(FAM[,1],FAM[,2])),,drop=FALSE]
+  ZZZ = scale(ZZZ,center=centerE)  # centering for E
+  write.table(cbind(FAM[,1:2],ZZZ),file=paste(fout,".cov",sep=""),row.names=FALSE,col.names=TRUE,quote=FALSE)
+
+  # cov for gxe
+  EZ = ZZZ  #
+  EE = EZ[,Enames,drop=FALSE];  # E (centered)
+
+  # lapprox
+  lapx = numeric(length(Enames)); names(lapx) = Enames
+  for(ee in 1:length(Enames)) lapx[ee] = .check.lapprox.ge(ZZZ,EE[,Enames[ee],drop=FALSE],y)
+  cat("\nlapprox: ")
+  print(lapx)
+
+  wlapx = c(1,lapx); wlapx[wlapx<1] = 1
+  if(maxal==ifelse(qua=="",1,3)*alo){
+    Maxal = pchisq(wlapx*qchisq(maxal,df=1,lower.tail=FALSE),df=1,lower.tail=FALSE)
+    Maxal[Maxal<1e-200] = 1e-200
+  }else{
+    Maxal = rep(maxal,length(Enames)+1)
   }
 
 
-  mean(crr)
+  ASSC = list()
+  ASSCGE = ASSCGEa = list(); ASSCGE[[1]] = ASSCGEa[[1]] = NULL; gxeparam2 = ""
+  aE0 = rep(TRUE,1+length(Enames))
+  if(qua==""){
 
-  matplot(cbind(colMeans(Op),colMeans(Oe),colMeans(Oe+2*GDF)),type="l")
+    # gxe (0: marginal)
+    nZ = ncol(ZZZ);
+    iZall = 1:(1+2*nZ)
 
-  #matplot(t(Op),type="l")
-  #matplot(t(Oe+2*GDF),type="l")
+    for(ee in 0:length(Enames)){
+      gxeparam = ""
+      if(ee>0){
+        Ename = Enames[ee]
+        iEinZ = which(colnames(ZZZ)==Ename)
+        gxeparam = paste0(" --interaction --tests ",(nZ+1)," --parameters ",paste(iZall[c(1+(1:nZ),1+nZ+iEinZ)],collapse=","))
+        if(AsscGE) gxeparam2 = paste0(" --interaction --parameters ",paste(c(1,iZall[c(1+(1:nZ),1+nZ+iEinZ)]),collapse=","))
+      }
+
+      system(paste(plink," --allow-no-sex  --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --logistic --covar ",fout,".cov --hide-covar --pfilter ",Maxal[ee+1]," --maf ",maf," --hwe ",hwe," --geno ",geno," --out ",fout,ee," --keep ",fout,".keep.fam",gxeparam,sep=""))
+      aE0[ee+1] = ( nrow(read.table(paste0(fout,ee,".assoc.logistic"),header=TRUE))>0 )
+      if(Assc) ASSC[[ee+1]] = read.table(paste0(fout,ee,".assoc.logistic"),header=TRUE)
+      if(AsscGE){
+        system(paste(plink," --allow-no-sex  --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --logistic --covar ",fout,".cov --maf ",maf," --hwe ",hwe," --geno ",geno," --out ",fout,ee,"__1"," --keep ",fout,".keep.fam",gxeparam,sep=""))
+        ASSCGE[[ee+1]] = read.table(paste0(fout,ee,"__1.assoc.logistic"),header=TRUE)
+        system(paste(plink," --allow-no-sex  --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --logistic --covar ",fout,".cov --maf ",maf," --hwe ",hwe," --geno ",geno," --out ",fout,ee,"__2"," --keep ",fout,".keep.fam",gxeparam2,sep=""))
+        ASSCGEa[[ee+1]] = read.table(paste0(fout,ee,"__2.assoc.logistic"),header=TRUE)
+      }
+      system(paste(plink," --recodeA --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --extract ",fout,ee,".assoc.logistic --out ",fout,ee," --keep ",fout,".keep.fam",sep=""))
+      system(paste(plink," --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --extract ",fout,ee,".assoc.logistic --freq --allow-no-sex --out ",fout,ee," --keep ",fout,".keep.fam",sep=""))
+    }
 
 
+  }else{
+
+    # gxe (0: marginal)
+    nZ = ncol(ZZZ);
+    iZall = 1:(1+2*nZ)
+    for(ee in 0:length(Enames)){
+      gxeparam = ""
+      if(ee>0){
+        Ename = Enames[ee]
+        iEinZ = which(colnames(ZZZ)==Ename)
+        gxeparam = paste0(" --vif 10000 --interaction --tests ",(nZ+1)," --parameters ",paste(iZall[c(1+(1:nZ),1+nZ+iEinZ)],collapse=","))
+        if(AsscGE) gxeparam2 = paste0(" --vif 10000 --interaction --parameters ",paste(c(1,iZall[c(1+(1:nZ),1+nZ+iEinZ)]),collapse=","))
+      }
+
+      system(paste(plink," --allow-no-sex  --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --linear --covar ",fout,".cov --hide-covar --pfilter ",Maxal[ee+1]," --maf ",maf," --hwe ",hwe," --geno ",geno," --out ",fout,ee," --keep ",fout,".keep.fam",gxeparam,sep=""))
+      aE0[ee+1] = ( nrow(read.table(paste0(fout,ee,".assoc.linear"),header=TRUE))>0 )
+      if(Assc) ASSC[[ee+1]] = read.table(paste0(fout,ee,".assoc.linear"),header=TRUE)
+      if(AsscGE){
+        system(paste(plink," --allow-no-sex  --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --linear --covar ",fout,".cov --maf ",maf," --hwe ",hwe," --geno ",geno," --out ",fout,ee,"__1"," --keep ",fout,".keep.fam",gxeparam,sep=""))
+        ASSCGE[[ee+1]] = read.table(paste0(fout,ee,"__1.assoc.linear"),header=TRUE)
+        system(paste(plink," --allow-no-sex  --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --linear --covar ",fout,".cov --maf ",maf," --hwe ",hwe," --geno ",geno," --out ",fout,ee,"__2"," --keep ",fout,".keep.fam",gxeparam2,sep=""))
+        ASSCGEa[[ee+1]] = read.table(paste0(fout,ee,"__2.assoc.linear"),header=TRUE)
+      }
+      system(paste(plink," --recodeA --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --extract ",fout,ee,".assoc.linear --out ",fout,ee," --keep ",fout,".keep.fam",sep=""))
+      system(paste(plink," --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --extract ",fout,ee,".assoc.linear --freq --allow-no-sex --out ",fout,ee," --keep ",fout,".keep.fam",sep=""))
+    }
+
+
+  }
+
+
+  ScX = list(); XXE = NULL; XE = list();
+  for(ee in 0:length(Enames)){
+    if(!aE0[ee+1]){ ScX[[ee+1]] = matrix(0,0,3); next }
+    Qr = read.table(paste(fout,ee,".frq",sep=""),colClasses="character",header=TRUE)[,2:3]
+    write.table( Qr, file=paste(fout,ee,".list",sep=""),row.names=FALSE,col.names=FALSE,quote=FALSE )  # For --recode-allele
+    ncX = scan(paste(fout,ee,".raw",sep=""),"character",nlines=1)
+    D = matrix(scan(paste(fout,ee,".raw",sep=""),"character"),nrow=length(ncX))[,-1]; colnames(D) = paste(D[1,],D[2,]);
+    XE[[ee+1]] = (t(D[-(1:6),rownames(FAM),drop=FALSE])=="1") + 2*(t(D[-(1:6),rownames(FAM),drop=FALSE])=="2");
+    XE[[ee+1]] = apply(XE[[ee+1]],2, function(x)ifelse(is.na(x), mean(x, na.rm=TRUE), x))
+    if(ee>0) XE[[ee+1]] = EE[,ee]*XE[[ee+1]]
+    colnames(XE[[ee+1]]) = paste0(ee,ncX[-(1:6)])
+    rm(D); gc(); gc();
+    sncX = strsplit(ncX[-(1:6)],"_")
+    ScX[[ee+1]] = cbind(Qr[,1],sapply(sncX,tail,1))
+    XXE = cbind(XXE,XE[[ee+1]])
+  }
+
+
+  ST = .stee(y=y,X=XXE,Z=ZZZ,Tau=Tau,qua=qua,maxal=maxal,gamma=gamma,ll=ll,naive=naive,kk=kk,alo=alo,grr=1,Sgn=Sgn,alc=alc,pSum=NULL)
+
+  DataTr = list(y=y,XE=XE,Z=ZZZ)
+
+  CP = ST$Loss + 2*ST$sig2hato*ST$gdf
+  if( which.min(CP)==1 && omit.tail ){
+    ilim = which( diff(CP)<0 )[1]
+    if(ilim<length(ST$al)) CP[ 1:ilim ] = Inf
+  }
+  #matplot(-log10(ST$al),CP,type="l")
+  lopt = which(CP==min(CP),arr.ind=TRUE)[1,]
+  Aopt = which(ST$BA[-(1:length(ST$dz)),lopt[1],lopt[2]]!=0)
+  Bopt = ST$BA[-(1:length(ST$dz)),lopt[1],lopt[2]]
+  allZero = all(Bopt==0)
+  names(lopt) = c("opt.al.index","opt.tau.index")
+
+  pXXE = rep(0:(length(ScX)-1),sapply(ScX,nrow))
+
+  if(!is.null(testbed) & length(testbed)==1) testbed = paste0(testbed,c(".bed",".bim",".fam"))
+
+
+  detectSTEE = list()
+  
+  # Estimated predicted values at optimal al for training data
+  if(!allZero){
+ 
+    mustp = ZZZ%*%ST$BA[1+(1:ncol(ZZZ)),lopt[1],lopt[2]] + outer(rep(1,nrow(ZZZ)),ST$BA[1,lopt[1],lopt[2]])
+    for(ee in 0:(length(ScX)-1)){
+      ipXXE = which(pXXE==ee)
+      if(length(ipXXE)==0){ detectSTEE[[ee+1]]=NA; next }
+
+      system(paste(plink," --out ",trainfout,ee," --bed ",trainbed[1]," --bim ",trainbed[2]," --fam ",trainbed[3]," --recodeA --recode-allele ",fout,ee,".list --extract ",fout,ee,".frq"," --keep ",fout,".keep.fam",sep=""))
+      if(!is.null(testbed)){
+        system(paste(plink," --out ",testfout,ee," --bed ",testbed[1]," --bim ",testbed[2]," --fam ",testbed[3]," --recodeA --recode-allele ",fout,ee,".list --extract ",fout,ee,".frq",sep=""))
+      }
+
+      ncX0 = scan(paste(trainfout,ee,".raw",sep=""),"character",nlines=1)
+      XG0 = t(matrix(scan(paste(trainfout,ee,".raw",sep=""),"character"),nrow=length(ncX0))[,-1]); XG0[which(XG0[,6]=="-9"),6] = NA; rownames(XG0) = paste(XG0[,1],XG0[,2]); XG0 = XG0[rownames(FAM),,drop=FALSE];
+      XG = (XG0[,-(1:6),drop=FALSE]=="1") + 2*(XG0[,-(1:6),drop=FALSE]=="2")
+      XGE = apply(XG,2, function(x)ifelse(is.na(x), mean(x, na.rm=TRUE), x))
+      if(ee>0) XGE = EE[,ee]*XGE
+      mustp = mustp + XGE%*%Bopt[ipXXE];
+      detectSTEE[[ee+1]] = cbind(ScX[[ee+1]],Bopt[ipXXE])
+    }
+
+    names(detectSTEE) = c("marginal",Enames)
+
+  }else{
+
+      mustp = ZZZ%*%ST$BA[1+(1:ncol(ZZZ)),lopt[1],lopt[2]] + outer(rep(1,nrow(ZZZ)),ST$BA[1,lopt[1],lopt[2]])
+
+  }
+
+
+  ## test data
+  if(!is.null(testbed)){
+    FAMte = read.table(testbed[3]); rownames(FAMte) = paste(FAMte[,1],FAMte[,2])
+
+    if(!is.null(testphenofile)){
+       stopifnot(!is.null(phenoname))
+       Phen = read.table(testphenofile,header=TRUE,na.strings="-9"); rownames(Phen) = as.character(paste(Phen[,1],Phen[,2]))
+       cat(paste0("Reading phenotype ",phenoname," from ",testphenofile,"\n"))
+       FAMte[,6] = NA; iii = intersect(rownames(FAMte),rownames(Phen)); FAMte[iii,6] = Phen[iii,phenoname]
+    }
+
+    # Estimated predicted values at optimal al for test data
+    if(is.character(Zte)){  # plink .cov format
+      Zte = read.table(Zte,header=TRUE,na.strings="-9"); rownames(Zte) = as.character(paste(Zte[,1],Zte[,2])); Zte = as.matrix(Zte[,-(1:2),drop=FALSE]);
+      if(!is.null(Znames)){
+         cat( paste0( "Used covariates: ",paste(intersect(colnames(Zte),Znames),collapse=","),"\n" ) )
+         Zte = Zte[,intersect(colnames(Zte),Znames),drop=FALSE]
+      }
+    }
+    stopifnot(nrow(Zte)==nrow(FAMte))
+    if(centerE){
+      ZZZte = scale(Zte[rownames(FAMte),],center=attr(ZZZ,"scaled:center"))  # centering for E
+    }else{
+      ZZZte = Zte[rownames(FAMte),]
+    } 
+    EEte = ZZZte[,Enames,drop=FALSE];  # E (centered)
+
+
+    if(!allZero){
+      mustpte = ZZZte[,,drop=FALSE]%*%ST$BA[1+(1:ncol(ZZZte)),lopt[1],lopt[2]] + outer(rep(1,nrow(FAMte)),ST$BA[1,lopt[1],lopt[2]])
+
+      for(ee in 0:(length(ScX)-1)){
+        ipXXE = which(pXXE==ee)
+        if(length(ipXXE)==0) next
+
+        ncX0 = scan(paste(testfout,ee,".raw",sep=""),"character",nlines=1)
+        XG0te = t(matrix(scan(paste(testfout,ee,".raw",sep=""),"character"),nrow=length(ncX0))[,-1]); XG0te[which(XG0te[,6]=="-9"),6] = NA; rownames(XG0te) = paste(XG0te[,1],XG0te[,2]); XG0te = XG0te[rownames(FAMte),,drop=FALSE]
+        XGte = (XG0te[,-(1:6),drop=FALSE]=="1") + 2*(XG0te[,-(1:6),drop=FALSE]=="2")
+        XGEte = apply(XGte,2, function(x)ifelse(is.na(x), mean(x, na.rm=TRUE), x)); #print(dim(XG0te))
+        if(ee>0) XGEte = EEte[,ee]*XGEte
+        mustpte = mustpte + XGEte%*%Bopt[ipXXE];
+      }
+
+    }else{
+
+      mustpte = ZZZte[rownames(FAMte),,drop=FALSE]%*%ST$BA[1+(1:ncol(ZZZte)),lopt[1],lopt[2]] + outer(rep(1,nrow(ZZZte)),ST$BA[1,lopt[1],lopt[2]])
+
+    }
+
+  }
+
+  PE = cbind(FAM,mustp)
+  PEte = NULL
+  if(!is.null(testbed)) PEte = cbind(FAMte,mustpte)
+
+
+ return( list( Muhat=ST$Muhat,gdf=ST$gdf,sig2hat=ST$sig2hat,df=ST$df,al=ST$al,lopt=lopt,BA=ST$BA,Loss=ST$Loss,sig2hato=ST$sig2hato,CP=CP,PE=PE,PEte=PEte,nonzero=detectSTEE,tau=ST$tau,DataTr=DataTr,lapprox=lapx,ASSC=ASSC,ASSCGE=ASSCGE,ASSCGEa=ASSCGEa) )
 }
+
+
+
+
+
+
+
+
